#include "mta-helper.fx"

// Input texture containing the world rendered with a toon shader
texture gToonDiffuse : TOONSAMPLER;

// Setting for the screen size
float2 gScreenSize : SCREENSIZE;

// Edge detection settings
float gEdgeThickness : EDGETHICKNESS;

// Sketching settings
texture gSketchSampler : SKETCHSAMPLER;
float2 gSketchJitter : SKETCHJITTER;
bool bSketching : SKETCHING;
bool bGrayscaleSketching : GRAYSCALESKETCHING;

// How should the sketch effect respond to changes of brightness in the input scene
float SketchThreshold = 0.15;
float SketchBrightness = 0.333;
 
sampler SketchSampler = sampler_state
{
    Texture         = (gSketchSampler);
    MinFilter       = Linear;
    MagFilter       = Linear;
    MipFilter       = Linear;
};

sampler SceneSampler = sampler_state
{
    Texture         = (gToonDiffuse);
    MinFilter       = Linear;
    MagFilter       = Linear;
    MipFilter       = Linear;
};

struct VSInput
{
    float3 Position : POSITION0;
    float3 Normal : NORMAL0;
    float2 TexCoord : TEXCOORD0;
};

struct PSInput
{
    float4 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
};

float pseudoBinarize(float minValue, float maxValue, float value)
{
	if ( value < minValue )
		return 0;
	if ( value > maxValue )
		return 1;
		
	return value;
}

float grayscale(float4 pixel)
{
	return pixel.r * 0.2126 + pixel.g * 0.7152 + pixel.b * 0.0722;
}

float processEdge(float2 coords)
{
	float dxtex = 1.0 / gScreenSize.x * gEdgeThickness;
	float dytex = 1.0 / gScreenSize.y * gEdgeThickness;
	float pix[9];
	int k = -1;

	// read neighboring pixel intensities
	for ( int i = -1; i < 2; ++i )
	{
		for ( int j = -1; j < 2; ++j )
		{
			++k;
			float4 pixel = tex2D(SceneSampler, coords + float2(i*dxtex, j*dytex));
			pix[k] = grayscale(pixel);
		}
	}

	// average color differences around neighboring pixels
	float delta = ( 	abs(pix[1]-pix[7]) +
						abs(pix[5]-pix[3]) +
						abs(pix[0]-pix[8]) +
						abs(pix[2]-pix[6])
					) * 0.25;
		   
	return pseudoBinarize ( 0.25, 0.45, saturate(delta*3.5) );
}

float4 PixelShaderFunction(PSInput PS) : COLOR0
{
    // Look up the original color from the main scene.
    float4 pixel = tex2D(SceneSampler, PS.TexCoord);
	
	if ( bSketching )
	{
		// Adjust the scene color to remove very dark values and increase the contrast.
		float3 saturatedScene = saturate((pixel - SketchThreshold) * 2);
		
		// Look up into the sketch pattern overlay texture.
		float3 sketchPattern = tex2D(SketchSampler, PS.TexCoord + gSketchJitter);
	
		// Convert into negative color space, and combine the scene color with the
		// sketch pattern. We need to do this multiply in negative space to get good
		// looking results, because pencil sketching works by drawing black ink
		// over an initially white page, rather than adding light to an initially
		// black background as would be more common in computer graphics.
		float3 negativeSketch = (1 - saturatedScene) * (1 - sketchPattern);
		
		// Convert the result into a positive color space greyscale value.
		float sketchResult = dot(1 - negativeSketch, SketchBrightness);
		
		// Apply the sketch result to the main scene color.
		if ( bGrayscaleSketching )
			pixel = sketchResult;
		else
			pixel *= sketchResult;
	}
	
	pixel *= (1 - processEdge(PS.TexCoord));

    return float4(pixel.rgb, 1);
}

PSInput VertexShaderFunction(VSInput VS)
{
    PSInput PS = (PSInput)0;
    PS.Position = mul(float4(VS.Position, 1), gWorldViewProjection);
    PS.TexCoord = VS.TexCoord;
    return PS;
}

technique postprocess
{
	pass P0
	{
		VertexShader = compile vs_3_0 VertexShaderFunction();
		PixelShader = compile ps_3_0 PixelShaderFunction();
	}
}
